/**
@license
Copyright (c) 2017 Vaadin Ltd.
This program is available under Apache License Version 2.0, available at https://vaadin.com/license/
*/
import { Grid as Grid$0 } from './vaadin-grid-scroller.js';

import { addListener } from '@polymer/polymer/lib/utils/gestures.js';
import { microTask, timeOut } from '@polymer/polymer/lib/utils/async.js';
import { PolymerElement } from '@polymer/polymer/polymer-element.js';
import { Base } from '@polymer/polymer/polymer-legacy.js';
import { FlattenedNodesObserver } from '@polymer/polymer/lib/utils/flattened-nodes-observer.js';
import { Debouncer } from '@polymer/polymer/lib/utils/debounce.js';
import { flush } from '@polymer/polymer/lib/utils/flush.js';
export const Grid = Grid$0 || {};

/**
 * @polymerMixin
 * TODO: remove this once the Chrome bug gets fixed: https://bugs.chromium.org/p/chromium/issues/detail?id=817499
 * There is a "limit" of 25 mixins allowed to apply, and Chrome crashes when exceeding this number
 */
Grid$0.CombinedMixin = superClass => class CombinedMixin extends superClass {

  static get properties() {
    return {
      /**
       * The item user has last interacted with. Turns to `null` after user deactivates
       * the item by re-interacting with the currently active item.
       */
      activeItem: {
        type: Object,
        notify: true,
        value: null
      },

      /**
       * An array containing the items which will be stamped to the column template
       * instances.
       */
      items: Array,

      _filters: {
        type: Array,
        value: function() {
          return [];
        }
      },

      /**
       * An array containing references to items with open row details.
       */
      detailsOpenedItems: {
        type: Array,
        value: function() {
          return [];
        }
      },

      _rowDetailsTemplate: Object,

      _detailsCells: {
        type: Array,
      },

      /**
       * When `true`, all `<vaadin-grid-sorter>` are applied for sorting.
       */
      multiSort: {
        type: Boolean,
        value: false
      },

      _sorters: {
        type: Array,
        value: function() {
          return [];
        }
      },

      _previousSorters: {
        type: Array,
        value: function() {
          return [];
        }
      },

      /**
       * An array that contains the selected items.
       */
      selectedItems: {
        type: Object,
        notify: true,
        value: () => []
      }
    };
  }

  static get observers() {
    return [
      '_a11yUpdateGridSize(size, _columnTree, _columnTree.*)',
      '_detailsOpenedItemsChanged(detailsOpenedItems.*, dataProvider, _rowDetailsTemplate, _physicalCountVal)',
      '_itemsChanged(items, items.*, isAttached)',
      '_rowDetailsTemplateChanged(_rowDetailsTemplate)',
      '_selectedItemsChanged(selectedItems.*)'
    ];
  }

  ready() {
    super.ready();

    const scroller = this.$.scroller;
    addListener(scroller, 'track', this._onHeaderTrack.bind(this));

    // Disallow scrolling while resizing
    scroller.addEventListener('touchmove', e => scroller.hasAttribute('column-resizing') && e.preventDefault());

    // Disable contextmenu on any resize separator.
    scroller.addEventListener('contextmenu', e => e.target.getAttribute('part') == 'resize-handle' && e.preventDefault());

    // Disable native cell focus when resizing
    scroller.addEventListener('mousedown', e => e.target.getAttribute('part') === 'resize-handle' && e.preventDefault());

    this.$.scroller.addEventListener('click', this._onClick.bind(this));

    this._addNodeObserver();

    this.addEventListener('filter-changed', this._filterChanged.bind(this));

    this.addEventListener('sorter-changed', this._onSorterChanged);

    this.addEventListener('cell-activate', this._activateItem.bind(this));

    // With Polymer 2 & shady the 'sorter-changed' listener isn't guaranteed to be registered
    // before child <vaadin-grid-sorter>'s upgrade and fire the events. The following
    // makes sure that 'sorter-changed' is fired for all <vaadin-grid-sorter> elements
    // after this (<vaadin-grid>) is ready (and the listeners are active).
    if (!undefined.useNativeShadow) {
      microTask.run(() => {
        const sorters = this.querySelectorAll('vaadin-grid-sorter');
        Array.from(sorters).forEach((sorter) => {
          // Don't try to fire if the sorter hasn't been upgraded yet
          if (sorter instanceof PolymerElement) {
            sorter.dispatchEvent(new CustomEvent('sorter-changed', {bubbles: true, composed: true}));
          }
        });
      });
    }
  }

  _a11yGetHeaderRowCount(_columnTree) {
    return _columnTree.filter(level => level.some(col => col._headerTemplate)).length;
  }

  _a11yGetFooterRowCount(_columnTree) {
    return _columnTree.filter(level => level.some(col => col._headerTemplate)).length;
  }

  _a11yUpdateGridSize(size, _columnTree) {
    if (size === undefined || _columnTree === undefined) {
      return;
    }

    const bodyColumns = _columnTree[_columnTree.length - 1];
    this.$.table.setAttribute(
      'aria-rowcount',
      size + this._a11yGetHeaderRowCount(_columnTree) + this._a11yGetFooterRowCount(_columnTree)
    );
    this.$.table.setAttribute('aria-colcount', bodyColumns && bodyColumns.length || 0);

    this._a11yUpdateHeaderRows();
    this._a11yUpdateFooterRows();
  }

  _a11yUpdateHeaderRows() {
    Array.from(this.$.header.children).forEach((headerRow, index) =>
      headerRow.setAttribute('aria-rowindex', index + 1)
    );
  }

  _a11yUpdateFooterRows() {
    Array.from(this.$.footer.children).forEach((footerRow, index) =>
      footerRow.setAttribute(
        'aria-rowindex',
        this._a11yGetHeaderRowCount(this._columnTree) + this.size + index + 1
      )
    );
  }

  _a11yUpdateRowRowindex(row, index) {
    row.setAttribute('aria-rowindex', index + this._a11yGetHeaderRowCount(this._columnTree) + 1);
  }

  _a11yUpdateRowSelected(row, selected) {
    // Jaws reads selection only for rows, NVDA only for cells
    row.setAttribute('aria-selected', Boolean(selected));
    Array.from(row.children).forEach(cell =>
      cell.setAttribute('aria-selected', Boolean(selected))
    );
  }

  _a11yUpdateRowLevel(row, level) {
    Array.from(row.children).forEach(cell =>
      // aria-level indexing starts from 1
      cell.setAttribute('aria-level', level + 1)
    );
    row.setAttribute('aria-level', level + 1);
  }

  _a11yUpdateRowDetailsOpened(row, detailsOpened) {
    Array.from(row.children).forEach(cell => {
      if (typeof detailsOpened === 'boolean') {
        cell.setAttribute('aria-expanded', detailsOpened);
      } else {
        if (cell.hasAttribute('aria-expanded')) {
          cell.removeAttribute('aria-expanded');
        }
      }
    });
  }

  _a11ySetRowDetailsCell(row, detailsCell) {
    Array.from(row.children).forEach(cell => {
      if (cell !== detailsCell) {
        cell.setAttribute('aria-controls', detailsCell.id);
      }
    });
  }

  _a11yUpdateCellColspan(cell, colspan) {
    cell.setAttribute('aria-colspan', Number(colspan));
  }

  _a11yUpdateSorters() {
    Array.from(this.querySelectorAll('vaadin-grid-sorter')).forEach(sorter => {
      let cellContent = sorter.parentNode;
      while (cellContent && cellContent.localName !== 'vaadin-grid-cell-content') {
        cellContent = cellContent.parentNode;
      }
      if (cellContent) {
        const cell = cellContent.assignedSlot.parentNode;
        cell.setAttribute('aria-sort', {
          'asc': 'ascending',
          'desc': 'descending'
        }[String(sorter.direction)] || 'none');
      }
    });
  }

  _activateItem(e) {
    const model = e.detail.model;
    const clickedItem = model ? model.item : null;

    if (clickedItem) {
      this.activeItem = !this._itemsEqual(this.activeItem, clickedItem) ? clickedItem : null;
    }
  }

  _itemsChanged(items, splices, isAttached) {
    if (!isAttached) {
      return;
    }
    if (!Array.isArray(items)) {
      if (items === undefined || items === null) {
        this.size = 0;
      }
      if (this.dataProvider === this._arrayDataProvider) {
        this.dataProvider = undefined;
      }
      return;
    }

    this.size = items.length;
    this.dataProvider = this.dataProvider || this._arrayDataProvider;
    this.clearCache();
    this._ensureFirstPageLoaded();
  }

  _arrayDataProvider(opts, cb) {
    let items = (Array.isArray(this.items) ? this.items : []).slice(0);

    if (this._filters && this._checkPaths(this._filters, 'filtering', items)) {
      items = this._filter(items);
    }

    this.size = items.length;

    if (opts.sortOrders.length && this._checkPaths(this._sorters, 'sorting', items)) {
      items = items.sort(this._multiSort.bind(this));
    }

    const start = opts.page * opts.pageSize;
    const end = start + opts.pageSize;
    const slice = items.slice(start, end);
    cb(slice, items.length);
  }

  /**
   * Check array of filters/sorters for paths validity, console.warn invalid items
   * @param {Array}  arrayToCheck The array of filters/sorters to check
   * @param {string} action       The name of action to include in warning (filtering, sorting)
   * @param {Array}  items
   */
  _checkPaths(arrayToCheck, action, items) {
    if (!items.length) {
      return false;
    }

    let result = true;

    for (var i in arrayToCheck) {
      const path = arrayToCheck[i].path;

      // skip simple paths
      if (!path || path.indexOf('.') === -1) {
        continue;
      }

      const parentProperty = path.replace(/\.[^\.]*$/, ''); // a.b.c -> a.b
      if (Base.get(parentProperty, items[0]) === undefined) {
        console.warn(`Path "${path}" used for ${action} does not exist in all of the items, ${action} is disabled.`);
        result = false;
      }
    }

    return result;
  }

  _multiSort(a, b) {
    return this._sorters.map(sort => {
      if (sort.direction === 'asc') {
        return this._compare(Base.get(sort.path, a), Base.get(sort.path, b));
      } else if (sort.direction === 'desc') {
        return this._compare(Base.get(sort.path, b), Base.get(sort.path, a));
      }
      return 0;
    }).reduce((p, n) => {
      return p ? p : n;
    }, 0);
  }

  _normalizeEmptyValue(value) {
    if ([undefined, null].indexOf(value) >= 0) {
      return '';
    } else if (isNaN(value)) {
      return value.toString();
    } else {
      return value;
    }
  }

  _compare(a, b) {
    a = this._normalizeEmptyValue(a);
    b = this._normalizeEmptyValue(b);

    if (a < b) {
      return -1;
    }
    if (a > b) {
      return 1;
    }
    return 0;
  }

  _filter(items) {
    return items.filter((item, index) => {
      return this._filters.filter(filter => {
        const value = this._normalizeEmptyValue(Base.get(filter.path, item));
        return value.toString().toLowerCase().indexOf(filter.value.toString().toLowerCase()) === -1;
      }).length === 0;
    });
  }

  // we need to listen to click instead of tap because on mobile safari, the
  // document.activeElement has not been updated (focus has not been shifted)
  // yet at the point when tap event is being executed.
  _onClick(e) {
    if (e.defaultPrevented) {
      // Something has handled this click already, e. g., <vaadin-grid-sorter>
      return;
    }

    const path = e.composedPath();
    const cell = path[path.indexOf(this.$.table) - 3];
    if (!cell || cell.getAttribute('part').indexOf('details-cell') > -1) {
      return;
    }
    const cellContent = cell._content;

    const activeElement = this.getRootNode().activeElement;
    const cellContentHasFocus = cellContent.contains(activeElement) &&
      // MSIE bug: flex children receive focus. Make type & attributes check.
      (!this._ie || this._isFocusable(activeElement));
    if (!cellContentHasFocus && !this._isFocusable(e.target)) {
      this.dispatchEvent(new CustomEvent('cell-activate', {detail: {model: cell._instance}}));
    }
  }

  _isFocusable(target) {
    if (!target.parentNode) {
      return false;
    }
    const focusables = Array.from(target.parentNode
      .querySelectorAll('[tabindex], button, input, select, textarea, object, iframe, label, a[href], area[href]'))
      .filter(element => element.getAttribute('part') !== 'cell body-cell');

    const isFocusableElement = focusables.indexOf(target) !== -1;
    return !target.disabled && isFocusableElement;
  }

  _onHeaderTrack(e) {
    const handle = e.target;
    if (handle.getAttribute('part') === 'resize-handle') {
      const cell = handle.parentElement;
      let column = cell._column;

      this._toggleAttribute('column-resizing', true, this.$.scroller);

      // Get the target column to resize
      if (column.localName === 'vaadin-grid-column-group') {
        column = Array.prototype.slice.call(column._childColumns, 0)
          .sort(function(a, b) {
            return a._order - b._order;
          })
          .filter(function(column) {
            return !column.hidden;
          }).pop();
      }

      const columnRowCells = Array.from(this.$.header.querySelectorAll('[part~="row"]:last-child [part~="cell"]'));
      var targetCell = columnRowCells.filter(cell => cell._column === column)[0];
      // Resize the target column
      if (targetCell.offsetWidth) {
        var style = window.getComputedStyle(targetCell);
        var minWidth = 10 + parseInt(style.paddingLeft) + parseInt(style.paddingRight) + parseInt(style.borderLeftWidth)
          + parseInt(style.borderRightWidth) + parseInt(style.marginLeft) + parseInt(style.marginRight);
        column.width = Math.max(minWidth, targetCell.offsetWidth + e.detail.x - targetCell.getBoundingClientRect().right) + 'px';
        column.flexGrow = 0;
      }
      // Fix width and flex-grow for all preceding columns
      columnRowCells
        .sort(
          function(a, b) {
            return a._column._order - b._column._order;
          })
        .forEach(function(cell, index, array) {
          if (index < array.indexOf(targetCell)) {
            cell._column.width = cell.offsetWidth + 'px';
            cell._column.flexGrow = 0;
          }
        });

      if (e.detail.state === 'end') {
        this._toggleAttribute('column-resizing', false, this.$.scroller);
      }

      // Notify resize
      this._resizeHandler();
    }
  }

  _hasColumnGroups(columns) {
    for (let i = 0; i < columns.length; i++) {
      if (columns[i].localName === 'vaadin-grid-column-group') {
        return true;
      }
    }

    return false;
  }

  _getChildColumns(el) {
    return FlattenedNodesObserver.getFlattenedNodes(el).filter(this._isColumnElement);
  }

  _flattenColumnGroups(columns) {
    return columns.map(col => {
      if (col.localName === 'vaadin-grid-column-group') {
        return this._getChildColumns(col);
      } else {
        return [col];
      }
    }).reduce((prev, curr) => {
      return prev.concat(curr);
    }, []);
  }

  _getColumnTree() {
    var rootColumns = FlattenedNodesObserver.getFlattenedNodes(this).filter(this._isColumnElement);
    var _columnTree = [];

    for (var c = rootColumns; ;) {
      _columnTree.push(c);
      if (!this._hasColumnGroups(c)) {
        break;
      }
      c = this._flattenColumnGroups(c);
    }

    return _columnTree;
  }

  _updateColumnTree() {
    var columnTree = this._getColumnTree();
    if (!this._arrayEquals(columnTree, this._columnTree)) {
      this._columnTree = columnTree;
    }
  }

  _addNodeObserver() {
    this._observer = new FlattenedNodesObserver(this, info => {

      const rowDetailsTemplate = info.addedNodes.filter(n => n.localName === 'template' && n.classList.contains('row-details'))[0];
      if (rowDetailsTemplate && this._rowDetailsTemplate !== rowDetailsTemplate) {
        this._rowDetailsTemplate = rowDetailsTemplate;
      }

      if (info.addedNodes.filter(this._isColumnElement).length > 0 ||
        info.removedNodes.filter(this._isColumnElement).length > 0) {
        this._updateColumnTree();
      }

      this._debouncerCheckImports = Debouncer.debounce(
        this._debouncerCheckImports,
        timeOut.after(2000),
        this._checkImports.bind(this));

      this._ensureFirstPageLoaded();
    });
  }

  _arrayEquals(arr1, arr2) {
    if (!arr1 || !arr2 || arr1.length != arr2.length) {
      return false;
    }

    for (var i = 0, l = arr1.length; i < l; i++) {
      // Check if we have nested arrays
      if (arr1[i] instanceof Array && arr2[i] instanceof Array) {
        // recurse into the nested arrays
        if (!this._arrayEquals(arr1[i], arr2[i])) {
          return false;
        }
      } else if (arr1[i] != arr2[i]) {
        return false;
      }
    }
    return true;
  }

  _checkImports() {
    [
      'vaadin-grid-column-group',
      'vaadin-grid-filter',
      'vaadin-grid-tree-toggle',
      'vaadin-grid-selection-column',
      'vaadin-grid-sorter'
    ].forEach(elementName => {
      var element = this.querySelector(elementName);
      if (element && !(element instanceof PolymerElement)) {
        console.warn(`Make sure you have imported the required module for <${elementName}> element.`);
      }
    });
  }

  _updateLastColumn() {
    Array.from(this.shadowRoot.querySelectorAll('tr')).forEach(row => {
      Array.from(row.querySelectorAll('[part~="cell"]:not([part~="details-cell"])'))
        .sort((a, b) => {
          return a._column._order - b._column._order;
        }).forEach((cell, cellIndex, children) => {
          this._toggleAttribute('last-column', cellIndex === children.length - 1, cell);
        });
    });
  }

  _isColumnElement(node) {
    return node.nodeType === Node.ELEMENT_NODE && /\bcolumn\b/.test(node.localName);
  }

  _filterChanged(e) {
    if (this._filters.indexOf(e.target) === -1) {
      this._filters.push(e.target);
    }

    e.stopPropagation();

    if (this.dataProvider) {
      this.clearCache();
    }
  }

  _mapFilters() {
    return this._filters.map(filter => {
      return {
        path: filter.path,
        value: filter.value
      };
    });
  }

  _rowDetailsTemplateChanged(rowDetailsTemplate) {
    if (rowDetailsTemplate) {
      var templatizer = new Grid$0.Templatizer();
      templatizer._grid = this;
      templatizer.dataHost = this.dataHost;
      templatizer.template = rowDetailsTemplate;
      rowDetailsTemplate.templatizer = templatizer;

      if (this._columnTree) {
        // Only update the rows if the column tree has already been initialized
        Array.from(this.$.items.children).forEach(row => {
          // TODO: Only update if there's no details cell
          // Ensure the rows have a row details cell
          this._updateRow(row, this._columnTree[this._columnTree.length - 1]);
          // Clear any old template instances
          delete row.querySelector('[part~=details-cell]')._instance;
        });
      }

      this._update();
    }
  }

  _detailsOpenedItemsChanged(changeRecord, dataProvider, rowDetailsTemplate, physicalCount) {
    if (dataProvider === undefined || !rowDetailsTemplate) {
      return;
    }

    if (changeRecord.path === 'detailsOpenedItems.length') {
      // Let’s avoid duplicate work of both “.splices” and “.length” updates.
      return;
    }

    this._flushItemsDebouncer();

    if (this._lastDetailsOpenedItems && physicalCount) {
      this._physicalItems.forEach(this._toggleDetailsCell, this);
      this._render();
    }

    this._lastDetailsOpenedItems = this.detailsOpenedItems;
  }

  _configureDetailsCell(cell) {
    cell.setAttribute('part', 'cell details-cell');
    // Freeze the details cell, so that it does not scroll horizontally
    // with the normal cells. This way it looks less weird.
    this._toggleAttribute('frozen', true, cell);
  }

  _toggleDetailsCell(row, item) {
    const cell = row.querySelector('[part~="details-cell"]');
    const detailsHidden = !this._isDetailsOpened(item);
    if (cell && (!cell._instance || cell.hidden !== detailsHidden)) {
      cell.hidden = detailsHidden;
      if (detailsHidden) {
        row.style.removeProperty('padding-bottom');
      } else {
        if (!cell._instance) {
          // Stamp the template
          cell._instance = this._rowDetailsTemplate.templatizer.createInstance();
          cell._content.innerHTML = '';
          cell._content.appendChild(cell._instance.root);
          this._updateItem(row, item);
        }

        flush();
        row.style.setProperty('padding-bottom', `${cell.offsetHeight}px`);
      }
    }
  }

  _updateDetailsCellHeights() {
    Array.from(this.$.items.querySelectorAll('[part~="details-cell"]:not([hidden])')).forEach(cell => {
      cell.parentElement.style.setProperty('padding-bottom', `${cell.offsetHeight}px`);
    });
  }

  _isDetailsOpened(item) {
    return this.detailsOpenedItems && this._getItemIndexInArray(item, this.detailsOpenedItems) !== -1;
  }

  /**
   * Open the details row of a given item.
   */
  openItemDetails(item) {
    if (!this._isDetailsOpened(item)) {
      this.push('detailsOpenedItems', item);
    }
  }

  /**
   * Close the details row of a given item.
   */
  closeItemDetails(item) {
    if (this._isDetailsOpened(item)) {
      this.splice('detailsOpenedItems', this._getItemIndexInArray(item, this.detailsOpenedItems), 1);
    }
  }

  _detailsOpenedInstanceChangedCallback(instance, value) {
    if (super._detailsOpenedInstanceChangedCallback) {
      super._detailsOpenedInstanceChangedCallback(instance, value);
    }
    if (value) {
      this.openItemDetails(instance.item);
    } else {
      this.closeItemDetails(instance.item);
    }
  }

  _isSelected(item) {
    return this.selectedItems && this._getItemIndexInArray(item, this.selectedItems) > -1;
  }

  /**
   * Selects the given item.
   *
   * @method selectItem
   * @param {Object} item The item object
   */
  selectItem(item) {
    if (!this._isSelected(item)) {
      this.push('selectedItems', item);
    }
  }

  /**
   * Deselects the given item if it is already selected.
   *
   * @method deselect
   * @param {Object} item The item object
   */
  deselectItem(item) {
    const index = this._getItemIndexInArray(item, this.selectedItems);
    if (index > -1) {
      this.splice('selectedItems', index, 1);
    }
  }

  /**
   * Toggles the selected state of the given item.
   *
   * @method toggle
   * @param {Object} item The item object
   */
  _toggleItem(item) {
    const index = this._getItemIndexInArray(item, this.selectedItems);
    if (index === -1) {
      this.selectItem(item);
    } else {
      this.deselectItem(item);
    }
  }

  _selectedItemsChanged(e) {
    if (this.$.items.children.length && (e.path === 'selectedItems' || e.path === 'selectedItems.splices')) {
      Array.from(this.$.items.children).forEach(row => {
        this._toggleRowSelected(row, this._isSelected(row._item));
      });
    }
  }

  _selectedInstanceChangedCallback(instance, value) {
    if (super._selectedInstanceChangedCallback) {
      super._selectedInstanceChangedCallback(instance, value);
    }
    if (value) {
      this.selectItem(instance.item);
    } else {
      this.deselectItem(instance.item);
    }
  }

  _onSorterChanged(e) {
    const sorter = e.target;

    this._removeArrayItem(this._sorters, sorter);
    sorter._order = null;

    if (this.multiSort) {
      if (sorter.direction) {
        this._sorters.unshift(sorter);
      }

      this._sorters.forEach((sorter, index) => sorter._order = this._sorters.length > 1 ? index : null, this);
    } else {
      this._sorters.forEach(sorter => {
        sorter._order = null;
        sorter.direction = null;
      });

      if (sorter.direction) {
        this._sorters = [sorter];
      }
    }

    e.stopPropagation();

    if (this.dataProvider &&
      // No need to clear cache if sorters didn't change
      JSON.stringify(this._previousSorters) !== JSON.stringify(this._mapSorters())) {
      this.clearCache();
    }

    this._a11yUpdateSorters();

    this._previousSorters = this._mapSorters();
  }

  _mapSorters() {
    return this._sorters.map(sorter => {
      return {
        path: sorter.path,
        direction: sorter.direction
      };
    });
  }

  _removeArrayItem(array, item) {
    const index = array.indexOf(item);
    if (index > -1) {
      array.splice(index, 1);
    }
  }
};
